import os, time
import psutil
from collections import namedtuple
import system_info_aix
import whatap.util.process_util as process_util
import whatap.util.thread_util as thread_util
import whatap.util.logging_util as logging_util
import whatap.agent.conf.configure as conf
from StringIO import StringIO

filesystem_blacklist = [
    "procfs"
]
class _DiskPerf:
    def __init__(self):
        self.mountPoint = None
        self.deviceId = None
        self.fileSystem = None
        self.totalSpace = None
        self.usedSpace = None
        self.usedPercent = None
        self.freeSpace = None
        self.freePercent = None
        self.blksize = None
        self.readIops = 0
        self.writeIops = 0
        self.readBps = 0
        self.writeBps = 0
        self.ioPercent = 0
        self.queueLength = 0

    def __str__(self):
        buf = StringIO()
        buf.write('mountPoint=')
        buf.write(self.mountPoint) 
        buf.write('\ndeviceId=')
        buf.write(self.deviceId)
        buf.write('\nfileSystem=')
        buf.write(self.fileSystem)
        buf.write('\ntotalSpace=')
        buf.write(self.totalSpace)
        buf.write('\nusedSpace=')
        buf.write(self.usedSpace)
        buf.write('\nusedPercent=')
        buf.write(self.usedPercent)
        buf.write('\nfreeSpace=')
        buf.write(self.freeSpace)
        buf.write('\nfreePercent=')
        buf.write(self.freePercent)
        buf.write('\nself.blksize=')
        buf.write(self.blksize)
        buf.write('\nreadIops=')
        buf.write(self.readIops)
        buf.write('\nwriteIops=')
        buf.write(self.writeIops)
        buf.write('\nreadBps=')
        buf.write(self.readBps)
        buf.write('\nwriteBps=')
        buf.write(self.writeBps)
        buf.write('\nioPercent=')
        buf.write(self.ioPercent)
        buf.write('\nqueueLength=')
        buf.write(self.queueLength)

        return buf.getvalue()

def iterateLines(lines, skip=0):
    for i, l in enumerate(lines.split('\n')):
        l = l.strip()
        if i < skip or not l:
            continue
        
        yield l.split()

_mountPointCache = None
def parseFromMountPoint(mountPoint):
    global _mountPointCache
    if _mountPointCache is None:
        _mountPointCache = {}
        lspv = process_util.executeCmd('lspv')
        
        for l in lspv.split('\n'):
            l = l.strip()
            if not l:
                continue
            (deviceId, serial, vgid, stauts)= l.split()
            lspvlist = process_util.executeCmd('lsvg -l %s'%(vgid))
            for tokens in iterateLines(lspvlist, skip=2):
                if len(tokens) < 7:
                    continue
                lvname, lvtype, lvlps, lvpps, lvpvs, lvstate, mountpoint = tokens
                _mountPointCache[mountpoint] = deviceId
    return _mountPointCache.get(mountPoint)


_mountPointCacheEx = None
def parseFromMountPointEx(mountPoint):
    global _mountPointCacheEx
    if _mountPointCacheEx is None:
        _mountPointCacheEx = {}
        lspv = process_util.executeCmd('lspv')
        
        for l in lspv.split('\n'):
            l = l.strip()
            if not l or len(l.split()) < 4:
                continue
            (deviceId, serial, vgid, stauts)= l.split()
            if 'none' == vgid.lower():
                continue
            lspvlist = process_util.executeCmd('lsvg -l %s'%(vgid))
            for tokens in iterateLines(lspvlist, skip=2):
                if len(tokens) < 7:
                    continue
                lvname, lvtype, lvlps, lvpps, lvpvs, lvstate, mountpoint = tokens
                if 'N/A' == mountpoint:
                    continue
                if mountpoint not in _mountPointCacheEx:
                    _mountPointCacheEx[mountpoint] = []    
                _mountPointCacheEx[mountpoint].append(deviceId)    
    return _mountPointCacheEx.get(mountPoint)

sdiskio = namedtuple('sdiskio', ['read_count', 'write_count',
                                 'read_bytes', 'write_bytes',
                                 'read_time', 'write_time', 'time'])

class _DiskIoPerfCache:
    def __init__(self, iocounters=None):
        self.iocounters= iocounters
        self.timestamp = time.time()

    def contains(self, mountPoint):
        deviceIds = parseFromMountPointEx(mountPoint)
        deviceFound = False
        if not deviceIds:
            return deviceFound
        for deviceId in deviceIds:
            deviceFound = deviceFound or deviceId in self.iocounters
        return deviceFound
    
    def get(self, mountPoint):
        deviceIds = parseFromMountPointEx(mountPoint)
        iotime = 0
        read_count = 0
        write_count = 0
        read_bytes = 0
        write_bytes = 0
        read_time = 0
        write_time = 0

        for deviceId in deviceIds:
            iocounter = self.iocounters.get(deviceId)
            iotime += iocounter.time
            read_count += iocounter.read_count
            write_count += iocounter.write_count
            read_bytes += iocounter.read_bytes
            write_bytes += iocounter.write_bytes
            read_time += iocounter.read_time
            write_time += iocounter.write_time
        iotime /= len(deviceIds)
        
        return sdiskio(time=iotime, read_count=read_count, write_count=write_count, read_bytes=read_bytes, write_bytes=write_bytes, read_time=read_time, write_time=write_time)

iocountersLastTime = _DiskIoPerfCache(psutil.disk_io_counters(perdisk=True))

def populate(mountPoint, callback, iocounters, iocountersLastTime):
    
    if not iocounters.contains(mountPoint) or not iocountersLastTime.contains(mountPoint):
        return
    iocounter = iocounters.get(mountPoint)
    iocounterLasttime =  iocountersLastTime.get(mountPoint)
    timediff = float(iocounters.timestamp - iocountersLastTime.timestamp)

    ioutil = float(100) * float(iocounter.time - iocounterLasttime.time) / float(system_info_aix.clk_tck) / timediff
    iopsRead = float(iocounter.read_count - iocounterLasttime.read_count) / timediff
    iopsWrite = float(iocounter.write_count - iocounterLasttime.write_count) / timediff
    bpsRead = float(iocounter.read_bytes - iocounterLasttime.read_bytes) /timediff
    bpsWrite = float(iocounter.write_bytes - iocounterLasttime.write_bytes) /timediff
    callback(ioutil, iopsRead,iopsWrite, bpsRead, bpsWrite )

def populateIoPerf(diskQuotas):
    global iocountersLastTime

    iocounters = _DiskIoPerfCache(psutil.disk_io_counters(perdisk=True))
    diskQuotasThisTime = []
    for diskQuota in diskQuotas:
        diskQuotasThisTime.append(diskQuota)
        mountPoint = diskQuota.mountPoint

        if not iocounters.contains(mountPoint):
            continue
        
        def setDiskIo(ioPercent, readiops, writeiops, readbps, writebps):
            diskQuota.readIops = readiops
            diskQuota.writeIops = writeiops
            diskQuota.readBps = readbps
            diskQuota.writeBps = writebps
            diskQuota.ioPercent = ioPercent


        populate(mountPoint, setDiskIo, iocounters, iocountersLastTime )
    
    iocountersLastTime = iocounters 

    return diskQuotasThisTime

def updateDiskPerf():
    global lastDiskPerf
    diskQuotas = []
    for i, part in enumerate(psutil.disk_partitions(all=False)):
        if part.fstype in filesystem_blacklist:
            continue
        p = _DiskPerf()
        p.mountPoint = part.mountpoint
        if not p.mountPoint:
            p.mountPoint = part.device
        p.deviceId = part.device
        p.fileSystem = part.fstype
        f_bsize, f_frsize, f_blocks, f_bfree, f_bavail, f_files, f_ffree, f_favail, f_flag, f_namemax = os.statvfs(p.mountPoint)
        p.totalSpace = int(f_frsize * f_blocks)
        p.usedSpace = int(f_frsize * (f_blocks - f_bfree))
        p.usedPercent = float(100.0 * float(f_blocks - f_bfree) / float(f_blocks))
        p.freeSpace = p.totalSpace - p.usedSpace
        p.freePercent = float(100.0 * f_bfree / f_blocks)
        p.blksize = int(f_frsize)
        diskQuotas.append(p)
        
    lastDiskPerf= populateIoPerf(diskQuotas)

lastDiskPerf = None
def getDisk():
    global lastDiskPerf
    return lastDiskPerf

try:
    updateDiskPerf()    
    thread_util.asyncLoop(updateDiskPerf, conf.GetConfig().DiskstatInterval)
except Exception, e:
    logging_util.debug("disk perf aix error:", e)

def test():
    for a in getDisk():
        print a

if __name__ == '__main__':
    test()
